/* extension.js
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/* exported init */

/*******************************************************************************
 * Ver como se hace este:
 * https://github.com/GSConnect/gnome-shell-extension-gsconnect/tree/master/src
*********************************************************************************/

'use strict';

const GLib = imports.gi.GLib;
const Gio = imports.gi.Gio;
const GObject = imports.gi.GObject;
const Gtk = imports.gi.Gtk;
const St = imports.gi.St;
const Clutter = imports.gi.Clutter;
const Gettext = imports.gettext;

const Lang = imports.lang;

const Main = imports.ui.main;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;
const AggregateMenu = Main.panel.statusArea.aggregateMenu;

const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();

const Domain = Gettext.domain(Me.metadata.uuid);
const _ = Domain.gettext;
const ngettext = Domain.ngettext;

// Rutas
const START = '/sys/class/power_supply/BAT0/charge_control_start_threshold';
const END = '/sys/class/power_supply/BAT0/charge_control_end_threshold';


const ThinkpadThresholdIndicator = GObject.registerClass({
  GTypeName: 'ThinkpadThresholdIndicator',
}, class ThinkpadThresholdIndicator extends PanelMenu.SystemIndicator {
  _init() {
    super._init();
    
    this.settings = ExtensionUtils.getSettings('org.gnome.shell.extensions.thinkpad-battery-threshold');
    
    this._start_file = null;
    this._end_file = null;
    this._enabled = false;
    
    this._threshold_start = 75;
    this._threshold_end = 80;
    
    // Obtener de la configuración
    this._load_prefs();
    
    this._active_icon = Gio.icon_new_for_string(Me.path + "/threshold-active-symbolic.svg");
    this._inactive_icon = Gio.icon_new_for_string(Me.path + "/threshold-inactive-symbolic.svg");
    this._unknown_icon = Gio.icon_new_for_string(Me.path + "/threshold-unknown-symbolic.svg");
    
    // Indicator
    this._indicator = this._addIndicator();
    this._indicator.gicon = this._unknown_icon;
    this._indicator.visible = true;
    
    AggregateMenu._indicators.insert_child_at_index(this, 0);
    AggregateMenu._threshold = this;
    
    // Menu principal
    this._item = new PopupMenu.PopupSubMenuMenuItem(_('Battery Threshold'), true);
    this._item.icon.gicon = Gio.icon_new_for_string(Me.path + "/threshold-app-symbolic.svg");
    this._item.label.clutter_text.x_expand = true;
    this.menu.addMenuItem(this._item);
    
    // Indicador del estado actual
    this._item._current_status = new St.BoxLayout({
    	opacity: 128,
    	style_class: 'current-status',
    });
    this._current_status_icon = new St.Icon({
            icon_size: 16,
        });
    this._item._current_status.add_child(this._current_status_icon);
    this._current_status_label = new St.Label({
            y_align: Clutter.ActorAlign.CENTER,
            text: _('Loading...'),
        });
    this._item._current_status.add_child(this._current_status_label);
    this._item.actor.insert_child_below(this._item._current_status, this._item._triangleBin);
    
    // Buscar el indice actual del menu de bateria
    const menuItems = AggregateMenu.menu._getMenuItems();
    const powerMenuIndex = AggregateMenu._power ? menuItems.indexOf(AggregateMenu._power.menu) : -1;
    const menuIndex = powerMenuIndex > -1 ? powerMenuIndex : 4;
    // Poner ese menu despues del menu de bateria
    AggregateMenu.menu.addMenuItem(this.menu, menuIndex + 1);    
		
	if (this._available()) {
		// Menu de accion
		this._toggle = new PopupMenu.PopupMenuItem(_('Loading...'));
		this._toggle.connect('activate', () => {
			this._enabled ? this._set_threshold(0, 100) : this._set_threshold(this._threshold_start, this._threshold_end);
		});
		this._item.menu.addMenuItem(this._toggle);
		
		// Separador
		this._item.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
		
		// Menu de configuracion
		this._config = new PopupMenu.PopupMenuItem(_('Settings'));
		this._config.connect('activate', () => {
			ExtensionUtils.openPrefs();
		});
		this._item.menu.addMenuItem(this._config);
		
		this.settings.connect('changed', () => {
			log('Loading settings...');
			this._load_prefs();
			this._update_gui();
		});
		
		this._update_gui();
	} else {
		this._current_status_icon.gicon = this._unknown_icon;
		this._current_status_label.visible = false;
	}
  }
  
  _load_prefs() {
  	const indicator_mode = this.settings.get_string('show-indicator');
  	const threshold = this.settings.get_value('threshold').deep_unpack();
    const threshold_start = threshold[0];
    const threshold_end = threshold[1];
    if (this._enabled && (threshold_start != this._threshold_start || threshold_end != this._threshold_end)) {
    	this._set_threshold(threshold_start, threshold_end)
    }
    this._indicator_mode = indicator_mode;
    this._threshold_start = threshold_start;
    this._threshold_end = threshold_end;
  }
  
  /*
  	Establece los umbrales
  */
  _set_threshold(start, end) {
  	
  	// Verifico que sean números
  	if (!Number.isInteger(start) || !Number.isInteger(end)) {
  		log('Invalid threshold values (Invalid format)! [' + start + '/' + end + ']');
  		return;
  	}
  	// Verifico que el inicio sea menor que el final, que el inicio sea mayor o igual que cero y que el final sea menor o igual a 100
  	if (start >= end || start < 0 || end > 100) {
  		log('Invalid threshold values (Invalid values)! [' + start + '/' + end + ']');
  		return;
  	}
  	// Verifico que sean múltiplos de 5
  	if ((start % 5) != 0 || (end % 5) != 0) {
  		log('Invalid threshold values (Is not a multiple of 5)! [' + start + '/' + end + ']');
  		return;
  	}
  	// Verifico límites
  	if (start != 0 || end != 100) {
  		if (start < 40 || start > 95 || end < 45 || end > 100) {
  			log('Invalid threshold values (Invalid ranges)! [' + start + '/' + end + ']');
  			return;
  		}
  	}
  	
  	log('Setting threshold: ' + start + '/' + end);
  	let notify = _('Impossible to apply battery charge thresholds!!!');
  	if (this._can_write()) {
  		// Modo normal, el usuario tiene permisos
  		try {
  			// Reinicio antes de aplicar los cambios para evitar valores inválidos
  			// Ej:
  			//		Actualmente está configurado en 50-55 y se quiere poner en 80-85... Se escribe el inicio en 80 que
  			//		es mayor a los 55 de fin actuales y eso da error.
  			this._start_file.replace_contents('0', null, false, Gio.FileCreateFlags.NONE, null);
	  		this._end_file.replace_contents('100', null, false, Gio.FileCreateFlags.NONE, null);
	  		
  			this._start_file.replace_contents(start.toString(), null, false, Gio.FileCreateFlags.NONE, null);
	  		this._end_file.replace_contents(end.toString(), null, false, Gio.FileCreateFlags.NONE, null);
	  		
	  		if (start != 0 && end != 100) {
  		  		notify = _('Battery charge thresholds enabled!!!');
  		  	} else {
  		  		notify = _('Battery charge thresholds disabled!!!');
  		  	}
  		} catch (e) {
  			logError(e, '_set_threshold');
  		}
  	} else {
		// Modo root, el usuario no tiene permisos
  		try {
  			// Reinicio antes de aplicar los cambios para evitar valores inválidos
  			// Ej:
  			//		Actualmente está configurado en 50-55 y se quiere poner en 80-85... Se escribe el inicio en 80 que
  			//		es mayor a los 55 de fin actuales y eso da error.
			const root_command = `pkexec sh -c "echo 0 > ${START}; echo 100 > ${END}; echo ${start.toString()} > ${START}; echo ${end.toString()} > ${END}"`;
			GLib.spawn_command_line_async(root_command);
			
			if (start != 0 && end != 100) {
  		  		notify = _('Battery charge thresholds enabled!!!');
  		  	} else {
  		  		notify = _('Battery charge thresholds disabled!!!');
  		  	}
  		} catch (e) {
  			logError(e, '_set_threshold');
  		}
  	}
  	
  	Main.notify(notify);
  }
  
  /*
  	Obtiene los valores actuales
  */
  _get_current_values() {
  	try {
  		const [, start_value,] = this._start_file.load_contents(null);
		const [, end_value,] = this._end_file.load_contents(null);
		return [parseInt(start_value), parseInt(end_value)];
	} catch (e) {
		logError(e, '_get_current_values');
		return [0, 0];
	} 
  }
  
  /*
  	Actualiza la interfaz
  */
  _update_gui() {
  	//log('_update_gui');
  	const [start, end] = this._get_current_values();
  	this._enabled = start != 0 || end != 100;
	this._toggle.label.text = this._enabled ? _('Disable threshold') : _('Enable threshold');
  	this._indicator.visible = (this._enabled || this._indicator_mode === 'always') && this._indicator_mode != 'never';
  	if (this._enabled) {
    	this._indicator.gicon = this._active_icon;
		this._current_status_icon.gicon = this._active_icon;
  	} else {
  		this._indicator.gicon = this._inactive_icon;
		this._current_status_icon.gicon = this._inactive_icon;
	}
  	this._current_status_label.text = _(`${start}/${end} %`);
	this._current_status_label.visible = true;
  }
  
  /*
  	Verifica si está disponible
  */
  _available() {
  	const files_exist = GLib.file_test(START, GLib.FileTest.EXISTS) && GLib.file_test(START, GLib.FileTest.EXISTS);
  	
  	if (files_exist) {
  		this._start_file = Gio.File.new_for_path(START);
		this._end_file = Gio.File.new_for_path(END);
		
		// Veo si puedo leer los datos
		if (this._can_read()) {
			// Monitor de cambios
			this._startMonitor = this._start_file.monitor_file(Gio.FileMonitorFlags.NONE, null);
		    this._startMonitor.connect("changed", this._update_gui.bind(this));
		    this._endMonitor = this._end_file.monitor_file(Gio.FileMonitorFlags.NONE, null);
		    this._endMonitor.connect("changed", this._update_gui.bind(this));
		} else {
			return false
		}
  	}  	
  	return files_exist;
  }
  
  _can_read() {
  	const start_info = this._start_file.query_info('access::*', Gio.FileQueryInfoFlags.NONE, null);
  	const end_info = this._end_file.query_info('access::*', Gio.FileQueryInfoFlags.NONE, null);
  	const can_read_start = start_info.get_attribute_boolean('access::can-read');
  	const can_read_end = end_info.get_attribute_boolean('access::can-read');
  	return can_read_start && can_read_end;
  }
  
  _can_write() {
  	const start_info = this._start_file.query_info('access::*', Gio.FileQueryInfoFlags.NONE, null);
  	const end_info = this._end_file.query_info('access::*', Gio.FileQueryInfoFlags.NONE, null);
  	const can_write_start = start_info.get_attribute_boolean('access::can-write');
  	const can_write_end = end_info.get_attribute_boolean('access::can-write');
  	return can_write_start && can_write_end;
  }
  
  destroy() {
  	if (this._startMonitor) {
        this._startMonitor.cancel();
        this._startMonitor = null;
    }
    if (this._endMonitor) {
        this._endMonitor.cancel();
        this._endMonitor = null;
    }
    this._item.destroy();
    this.menu.destroy();
    delete AggregateMenu._threshold;
    delete this._start_file;
    delete this._end_file;
    super.destroy();
  }
});

var thresholdIndicator = null;

function init() {
	ExtensionUtils.initTranslations(Me.metadata.uuid);
}

function enable() {
  thresholdIndicator = new ThinkpadThresholdIndicator();
}

function disable() {
  thresholdIndicator.destroy();
  thresholdIndicator = null;
}
